/**
 *	@file	devAai5565.c
 *	@brief	RM5565 device support for aaiRecord
 *	$Revision: 21 $
 *	$Date:: 2022-11-24 11:03:23 +0900#$
 *	$Author: iitsuka $
 *
 *	@page devsup_aai Device Support for Reading memory
 *	リフレクティブメモリデータ読み出し用デバイスサポート
 *
 *	このデバイスサポートは、aai
 *	レコードタイプと組み合わせて使用する事で、
 *	リフレクティブメモリからのデータ読み出しを行うためのものである。@n
 *	Device Type 名は @c "aai5565"
 *
 *	@par Synopsis
 *	@code{.unparsed}
 *	record(aai, "pv-name") {
 *		field(DTYP, "aai5565")
 *		field(INP,  "#C<card_number> S @<offset>")
 *		field(FTVL, "<value_type>")
 *		field(NELM, "<num_of_elems>")
 *	}
 *	@endcode
 *	@b card_number
 *	@li The card number designated in argument @e card in @ref RM5565PciConfigure
 *
 *	@b offset
 *	@li Offset of RM5565 memory
 *
 *	@b value_type
 *	@li Field Type of Value
 *
 *	@b num_of_elems
 *	@li Number of elements
 *
 *	@par Usage
 *	@code{.unparsed}
 *	record(aai, "RM5565:DEV0:A000:R") {
 *		field(DTYP, "aai5565")
 *		field(INP,  "#C0 S @0xA000")
 *		field(FTVL, "LONG")
 *		field(NELM, "1")
 *			・
 *			・
 *			・
 *	}
 *	@endcode
 *	@c INP フィールドでカード番号/オフセットを指定する。@n
 *	@c "C" に続く数字(例では @c '0')でカード番号を、
 *	@c "@" に続く数字(例では @c '0xA000')でオフセットを指定しておく。@n
 *	指定できるオフセットは 0x0000～である。ちなみに VME版では 0x3000～であり、
 *	PCI版の 0x0000 と VME版の 0x3000 が同じアドレスになる。 @n
 *	オフセットの指定には 16 進表記以外にも 10 進表記や 8 進表記も使用可能。@n
 *	@code
 *		field(INP,  "#C0 S @40960") ... 10 進表記
 *	@endcode
 *	@code
 *		field(INP,  "#C0 S @0120000") ... 8 進表記
 *	@endcode
 *	このレコードをプロセスすると、リフレクティブメモリのオフセット 0xA000
 *	番地から 32 ビットデータを読み込み、@c VAL フィールドにセットする。
 *	@code
 *	$ caput RM5565:DEV0:A000:R.PROC 1
 *	Old : RM5565:DEV0:A000:R.PROC 0
 *	New : RM5565:DEV0:A000:R.PROC 1
 *	$ caget RM5565:DEV0:A000:R
 *	RM5565:DEV0:A000:R     0
 *	@endcode
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <alarm.h>
#include <dbAccess.h>
#include <dbEvent.h>
#include <dbFldTypes.h>
#include <dbScan.h>
#include <devSup.h>
#include <recSup.h>
#include <recGbl.h>
#include <cantProceed.h>
#include <epicsExport.h>
#include <errlog.h>

#include "aaiRecord.h"
#include "drv5565.h"

typedef	struct
{
	RM5565_t handle;
	epicsUInt32 offset;
	epicsUInt32 bytes;
} dpvt_t;

/**	add specific record
 *	@param[in,out]	pdbc	レコード情報へのポインタ
 *	@retval	0					正常終了
 *	@retval	S_dev_badInpType	INPフィールドのリンクタイプが不正
 *	@retval	S_dev_badCard		INPフィールドのカード番号指定が不正
 *	@retval	S_db_badField		INPフィールドのオフセット指定が不正
 */
static long add_record(dbCommon *pdbc)
{
	struct aaiRecord *prec = (struct aaiRecord *)pdbc;
	long status;

	if (prec->inp.type != VME_IO)
	{
		(void)errlogPrintf(__FILE__"(%d)[%s] invalid link-type(=%d)\n", __LINE__, prec->name, prec->inp.type);
		status = S_dev_badInpType;
	}
	else
	{
		RM5565_t handle;

		handle = drvRM5565_getHandle(prec->inp.value.vmeio.card);
		if (handle == NULL)
		{
			(void)errlogPrintf(__FILE__"(%d)[%s] card:%d is not installed\n", __LINE__, prec->name, prec->inp.value.vmeio.card);
			status = S_dev_badCard;
		}
		else
		{
			ptrdiff_t offset;
			char *endp;

			offset = (ptrdiff_t)strtol(prec->inp.value.vmeio.parm, &endp, 0);
			if (((*endp == '\0') || isspace(*endp)) && drvRM5565_validOffset(handle, offset))
			{
				dpvt_t *dpvt;

				dpvt = (dpvt_t *)mallocMustSucceed(sizeof(dpvt_t), "aai5565: init_record");
				dpvt->handle = handle;
				dpvt->offset = offset;
				dpvt->bytes = dbValueSize(prec->ftvl) * prec->nelm;
				prec->dpvt = dpvt;
				status = 0L;
			}
			else
			{
				(void)errlogPrintf(__FILE__"(%d)[%s] invalid param(=%s)\n", __LINE__, prec->name, prec->inp.value.vmeio.parm);
				status = S_db_badField;
			}
		}
	}
	return status;
}

/**	delete specific record
 *	@param[in,out]	pdbc	レコード情報へのポインタ
 *	@return	必ず0を返す
 */
static long del_record(dbCommon *pdbc)
{
	void *addr = pdbc->dpvt;

	pdbc->dpvt = NULL;
	free(addr);

	return 0L;
}

/**	initialize device processing
 *	@param[in]	after	before/after initializing record
 *						- @c 0 : before initializaing record
 *						- @c 1 : after initializing record
 *	@return	必ず0を返す
 */
static long init_aai(int after)
{
	if (after == 0)
	{
		static struct dsxt dsxt = {add_record, del_record};

		devExtend(&dsxt);
	}
	return 0L;
}

/**	initialize specific record
 *	@param[in,out]	prec	pointer to record information
 *	@return	必ず0を返す
 */
static long init_record(struct aaiRecord *prec)
{
	return 0L;
}

/**	read specific record
 *	@param[in,out]	prec	pointer to record information
 *	@return	必ず0を返す
 */
static long read_record(struct aaiRecord *prec)
{
	dpvt_t *dpvt = (dpvt_t *)prec->dpvt;

	if (dpvt != NULL)
	{
		if (drvRM5565_read(dpvt->handle, dpvt->offset, prec->bptr, dpvt->bytes) != 0)
		{
			recGblSetSevr(prec, COMM_ALARM, MAJOR_ALARM);
		}
		else
		{
			prec->nord = prec->nelm;
		}
	}
	return 0L;
}

/**	@cond		*/
static const struct {
	long number;
	DEVSUPFUN report;
	DEVSUPFUN init;
	DEVSUPFUN init_record;
	DEVSUPFUN get_ioint_info;
	DEVSUPFUN read_record;
} devAai5565 = {
	5,
	NULL,
	init_aai,
	init_record,
	NULL,
	read_record
};
epicsExportAddress(dset, devAai5565);
/**	@endcond	*/

