/**
 *	@file	devAao5565.c
 *	@brief	RM5565 device support for aaoRecord
 *	$Revision: 21 $
 *	$Date:: 2022-11-24 11:03:23 +0900#$
 *	$Author: iitsuka $
 *
 *	@page devsup_aao Device Support for Writing memory
 *	リフレクティブメモリデータ書き込み用デバイスサポート
 *
 *	このデバイスサポートは、aao
 *	レコードタイプと組み合わせて使用する事で、
 *	リフレクティブメモリへデータ書き込みを行うためのものである。@n
 *	Device Type 名は @c "aao5565"
 *
 *	@par Synopsis
 *	@code{.unparsed}
 *	record(aao, "pv-name") {
 *		field(DTYP, "aao5565")
 *		field(OUT,  "#C<card_number> S @<offset>[,<wait>]")
 *		field(FTVL, "<value_type>")
 *		field(VAL,  "<data>")
 *		field(NELM, "<num_of_elems>")
 *	}
 *	@endcode
 *	@b card_number
 *	@li The card number designated in argument @e card in @ref RM5565PciConfigure
 *
 *	@b offset
 *	@li offset of RM5565 memory
 *
 *	@b wait
 *	@li waiting time for completion of write
 *
 *	@b value_type
 *	@li Field Type of Value
 *
 *	@b num_of_elems
 *	@li Number of elements
 *
 *	@b data
 *	@li data to write
 *
 *	@par Usage
 *	@code{.unparsed}
 *	record(aao, "RM5565:DEV0:A000:W") {
 *		field(DTYP, "aao5565")
 *		field(OUT,  "#C0 S @0xA000,1.0")
 *		field(FTVL, "SHORT")
 *		field(NELM, "1")
 *			・
 *			・
 *			・
 *		field(VAL,  "<data>")
 *	}
 *	@endcode
 *	@c OUT フィールドでカード番号/オフセットを指定する。@n
 *	@c "C" に続く数字(例では @c '0')でカード番号を、
 *	@c "@" に続く数字(例では @c '0xA000')でオフセットを指定しておく。@n
 *	指定できるオフセットは 0x0000～である。ちなみに VME版では 0x3000～であり、
 *	PCI版の 0x0000 と VME版の 0x3000 が同じアドレスになる。 @n
 *	オフセットの指定には 16 進表記以外にも 10 進表記や 8 進表記も使用可能。@n
 *	@code
 *		field(OUT,  "#C0 S @40960") ... 10 進表記
 *	@endcode
 *	@code
 *		field(OUT,  "#C0 S @0120000") ... 8 進表記
 *	@endcode
 *	また、オプションとして書き込み完了待ちの最大時間を秒で指定可能 
 *	(例では @c '1.0')である。@n
 *
 *	書き込み待ちの処理は @c PRIO フィールドで指定された General Purpose Callback Task
 *	のコンテキストで非同期式デバイスサポート処理として実行され、
 *	@ref drvRM5565_waitOwnData を使用して完了までの待ち処理を行う。@n
 *	もし、最大待ち時間を過ぎても書き込みが完了しなかった 
 *	(OWN DATA ビットがセットされなかった)場合、
 *	レコードは COMM_ALARM/MINOR_ALARM 状態になる。@n
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <alarm.h>
#include <dbAccess.h>
#include <dbEvent.h>
#include <dbFldTypes.h>
#include <dbScan.h>
#include <devSup.h>
#include <recSup.h>
#include <recGbl.h>
#include <epicsThread.h>
#include <epicsMessageQueue.h>
#include <callback.h>
#include <cantProceed.h>
#include <epicsExport.h>
#include <errlog.h>

#include "aaoRecord.h"
#include "drv5565.h"

typedef	struct {
	RM5565_t handle;
	epicsUInt32 offset;
	epicsUInt32 bytes;
	epicsFloat64 wait;
	CALLBACK callback;
} dpvt_t;

/**	skip space
 *	@param[in]	str	string
 *	@return	空白スキップ後のポインタ @n
 *			空白文字のみだった場合には NULL を返す
 */
static const char *skip_space(const char *str)
{
	while (*str != '\0')
	{
		if (!isspace(*str))
		{
			return str;
		}
		str++;
	}
	return NULL;
}

/**	get parameters
 *	@param[in]	param		parameter string
 *	@param[out]	p_offset	pointer to a variable where offset parameter will be stored
 *	@param[out]	p_wait		pointer to a variable where wait parameter will be stroed
 *	@retval	0	OK
 *	@retval	-1	NG
 */
static long get_param(const char *param, epicsUInt32 *p_offset, epicsFloat64 *p_wait)
{
	epicsUInt32 offset;
	char *endp;
	long status;

	offset = strtol(param, &endp, 0);
	if (*endp == '\0')
	{
		*p_offset = offset;
		*p_wait = 0.0;
		status = 0L;
	}
	else
	{
		param = skip_space(endp);
		if ((param == NULL) || (*param != ','))
		{
			status = -1L;
		}
		else
		{
			epicsFloat64 wait;

			wait = strtod(++param, &endp);
			if (param == endp)
			{
				status = -1L;
			}
			else if ((*endp != '\0') && !isspace(*endp))
			{
				status = -1L;
			}
			else
			{
				*p_offset = offset;
				*p_wait = wait;
				status = 0L;
			}
		}
	}
	return status;
}

/**	polling OWN-DATA bit
 *	@param[in,out]	pCB		pointer to callback information
 */
static void poll_owndata(CALLBACK *pCB)
{
	dbCommon *prec;
	dpvt_t *dpvt;
	struct rset *prset;

	callbackGetUser(prec, pCB);
	assert(strcmp(prec->rdes->name, "aao") == 0);
	dpvt = (dpvt_t *)prec->dpvt;

	if (drvRM5565_waitOwnData(dpvt->handle, dpvt->wait) != 0)
	{
		recGblSetSevr(prec, COMM_ALARM, MINOR_ALARM);
	}
	prset = (struct rset *)prec->rset;
	dbScanLock(prec);
	(*prset->process)(prec);
	dbScanUnlock(prec);
}

/**	add specific record
 *	@param[in,out]	pdbc	レコード情報へのポインタ
 *	@retval	0					正常終了
 *	@retval	S_dev_badOutType	OUTフィールドのリンクタイプが不正
 *	@retval	S_dev_badCard		OUTフィールドのカード番号指定が不正
 *	@retval	S_db_badField		OUTフィールドの記述が不正
 */
static long add_record(dbCommon *pdbc)
{
	struct aaoRecord *prec = (struct aaoRecord *)pdbc;
	long status;

	if (prec->out.type != VME_IO)
	{
		(void)errlogPrintf(__FILE__"(%d)[%s] invalid link-type(=%d)\n", __LINE__, prec->name, prec->out.type);
		status = -1L;
	}
	else
	{
		RM5565_t handle;

		handle = drvRM5565_getHandle(prec->out.value.vmeio.card);
		if (handle == NULL)
		{
			(void)errlogPrintf(__FILE__"(%d)[%s] card:%d is not installed\n", __LINE__, prec->name, prec->out.value.vmeio.card);
			status = S_dev_badCard;
		}
		else
		{
			epicsUInt32 offset;
			epicsFloat64 wait;

			if ((get_param(prec->out.value.vmeio.parm, &offset, &wait) != 0L) || !drvRM5565_validOffset(handle, offset))
			{
				(void)errlogPrintf(__FILE__"(%d)[%s] invalid param(=%s)\n", __LINE__, prec->name, prec->out.value.vmeio.parm);
				status = S_db_badField;
			}
			else
			{
				dpvt_t *dpvt;

				dpvt = (dpvt_t *)mallocMustSucceed(sizeof(dpvt_t), "aao5565: init_record");
				dpvt->handle = handle;
				dpvt->offset = offset;
				dpvt->bytes = dbValueSize(prec->ftvl) * prec->nelm;
				dpvt->wait = wait;
				callbackSetCallback(poll_owndata, &dpvt->callback);
				callbackSetPriority(prec->prio, &dpvt->callback);
				callbackSetUser(prec, &dpvt->callback);

				prec->dpvt = dpvt;
				status = 0L;
			}
		}
	}
	return status;
}

/**	delete specific record
 *	@param[in,out]	pdbc	レコード情報へのポインタ
 *	@return	必ず0を返す
 */
static long del_record(dbCommon *pdbc)
{
	void *addr = pdbc->dpvt;

	pdbc->dpvt = NULL;
	free(addr);

	return 0L;
}

/**	initialize device processing
 *	@param[in]	after	before/after initializing record
 *						- @c 0 : before initializaing record
 *						- @c 1 : after initializing record
 *	@return	必ず0を返す
 */
static long init_aao(int after)
{
	if (after == 0)
	{
		static struct dsxt dsxt = {add_record, del_record};

		devExtend(&dsxt);
	}
	return 0L;
}

/**	initialize specific record
 *	@param[in,out]	prec	pointer to aaoRecord structure
 *	@return	必ず0を返す
 */
static long init_record(struct aaoRecord *prec)
{
	return 0L;
}

/**	write specific record
 *	@param[in,out]	prec	pointer to aaoRecord structure
 *	@return	必ず0を返す
 */
static long write_record(struct aaoRecord *prec)
{
	dpvt_t *dpvt = (dpvt_t *)prec->dpvt;

	if (dpvt != NULL)
	{
		if (prec->pact != FALSE)
		{
			prec->nord = prec->nelm;
		}
		else
		{
			if (dpvt->wait > 0.0)
			{
				drvRM5565_clearOwnData(dpvt->handle);
			}
			if (drvRM5565_write(dpvt->handle, dpvt->offset, prec->bptr, dpvt->bytes) != 0)
			{
				recGblSetSevr(prec, COMM_ALARM, MAJOR_ALARM);
			}
			if (dpvt->wait > 0.0)
			{
				prec->pact = TRUE;
				callbackRequest(&dpvt->callback);
			}
			else
			{
				prec->nord = prec->nelm;
			}
		}
	}
	return 0L;
}

/**	@cond		*/
static const struct {
	long number;
	DEVSUPFUN report;
	DEVSUPFUN init;
	DEVSUPFUN init_record;
	DEVSUPFUN get_ioint_info;
	DEVSUPFUN write_record;
} devAao5565 = {
	5,
	NULL,
	init_aao,
	init_record,
	NULL,
	write_record
};
epicsExportAddress(dset, devAao5565);
/**	@endcond	*/

