/**
 *	@file	devIntrx5565.c
 *	@brief	Device support for receiving network interrupt of RM5565
 *	$Revision: 21 $
 *	$Date:: 2022-11-24 11:03:23 +0900#$
 *	$Author: iitsuka $
 *
 *	@page devsup_intrx Device Support for Receiving Network-Interrupt
 *	Network-Interrupt 受信用デバイスサポート
 *
 *	このデバイスサポートは、@ref recsup_intrx "intrx"
 *	レコードタイプと組み合わせて使用する事で、
 *	Network-Interrupt の受信を行うためのものである。@n
 *	Device Type 名は @c "intrx5565"。
 *
 *	@par Synopsis
 *	@code{.unparsed}
 *	record(intrx, "pv-name") {
 *		field(DTYP, "intrx5565")
 *		field(SCAN, "I/O Intr")
 *		field(INP,  "#C<card_number> S<int_type> @")
 *		field(SNAM, "<subroutine_name>")
 *	}
 *	@endcode
 *	@b card_number
 *	@li The card number designated in argument @e card in @ref RM5565PciConfigure
 *
 *	@b int_type
 *	@li @c 1 ... Type 1 Network-Interrupt
 *	@li @c 2 ... Type 2 Network-Interrupt
 *	@li @c 3 ... Type 3 Network-Interrupt
 *	@li @c 4 ... Type 4 Network-Interrupt
 *
 *	@b subroutine_name
 *	@li Network-Interrupt 発生時に呼び出す関数名(省略可)
 *
 *	@par Usage
 *	@code{.unparsed}
 *	record(intrx, "RM5565:DEV0:INT1:RX") {
 *		field(DTYP, "intrx5565")
 *		field(SCAN, "I/O Intr")
 *		field(INP,  "#C0 S1 @")
 *		field(SNAM, "int1func")
 *			・
 *			・
 *			・
 *	}
 *	@endcode
 *	@c INP フィールドでカード番号/Network-Interrupt Type を指定する。@n
 *	@c "C" に続く数字 (例では @c '0') でカード番号を、
 *	@c "S" に続く数字 (例では @c '1') で Network-Interrupt Type を指定しておく。@n
 *	他ノードから自ノードに対して指定した Network-Interrupt 
 *	が送信されるとレコードがプロセスされる。@n
 *	その際、下記フィールドには以下の値がセットされる。@n
 *	Field | セットされる値
 *	:----:|:------------------------------------------
 *	@c NID| Network-Interrupt を送信したノードのID @n
 *	@c VAL| Network-Interrupt Data @n
 *	また、フィールド @c SNAM で指定されたサブルーチン(例では @c int1func)がコールされる。@n
 *	関数側の記述例は下記の通り。
 *	@code{.c}
 *	void int1func(struct intrxRecord *prec)
 *	{
 *			・
 *			・
 *			・
 *	}
 *	epicsRegisterFunction(int1func);
 *	@endcode
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <alarm.h>
#include <dbAccess.h>
#include <dbScan.h>
#include <recSup.h>
#include <devSup.h>
#include <recGbl.h>
#include <epicsExport.h>
#include <cantProceed.h>
#include <errlog.h>

#include "intrxRecord.h"
#include "drv5565.h"

/**
 *	@brief	intrx レコード内部情報構造体.
 */
typedef	struct
{
	RM5565_t handle;	/**< VMIVME-5565のハンドル								*/
	epicsUInt8 intId;	/**< レコードが対応するnetwork interruptのID (0 ~ 3)	*/
} intrx_dpvt_t;

/**	add specific record
 *	@param[in,out]	prec	pointer to intrxRecord structure
 *	@retval	0					正常終了
 *	@retval	S_dev_badInpType	INPフィールドのリンクタイプが不正
 *	@retval	S_dev_badCard		INPフィールドのカード番号指定が不正
 *	@retval	S_dev_badSignal		INPフィールドのシグナル番号指定が不正
 *
 *	- - -
 *	_INP_ フィールドで指定されたカード番号/Network Interrupt Type に対応する __intrx__ レコードの初期化を行う。@n
 *	@code
 *	record(intrx, "pv-name") {
 *	        ・
 *	        ・
 *	    field(INP, "#C0 S1 @")
 *	        ・
 *	        ・
 *	}
 *	@endcode
 *	@c "C" に続く数字 (ここでは @c '0') でカード番号が、
 *	@c "S" に続く数字 (ここでは @c '1') で Network Interrupt Type が指定される。
 *
 *	カード番号に対応するリフレクティブメモリのハンドル (@ref RM5565_t) を取得し、
 *	Network Interrupt Type とともに、@ref intrx_dpvt_t 構造体のそれぞれのメンバに格納する。@n
 *	値の格納された @ref intrx_dpvt_t は _DPVT_ フィールドにセットされる。@n
 *	@see	drvRM5565_getHandle
 */
static long add_record(dbCommon *pdbc)
{
	struct intrxRecord *prec = (struct intrxRecord *)pdbc;
	long status;

	if (prec->inp.type != VME_IO)
	{
		(void)errlogPrintf(__FILE__"(%d) Bad INP-type\n", __LINE__);
		status = S_dev_badInpType;
	}
	else
	{
		struct vmeio *pvme = &prec->inp.value.vmeio;
		RM5565_t handle;

		handle = drvRM5565_getHandle(pvme->card);
		if (handle == NULL)
		{
			(void)errlogPrintf(__FILE__"(%d) card#(=%d) is not installed\n", __LINE__, pvme->card);
			status = S_dev_badCard;
		}
		else if ((1 > pvme->signal) || (pvme->signal > 4))
		{
			(void)errlogPrintf(__FILE__"(%d) invalid signal#(=%d)\n", __LINE__, pvme->signal);
			status = S_dev_badSignal;
		}
		else
		{
			intrx_dpvt_t *dpvt = (intrx_dpvt_t *)mallocMustSucceed(sizeof(intrx_dpvt_t), "devIntrx5565: init_record");
			dpvt->handle = handle;
			dpvt->intId = (epicsUInt8)pvme->signal - 1u;
			prec->dpvt = dpvt;
			status = 0L;
		}
	}
	return status;
}

/**	delete specific record
 *	@param[in,out]	pdbc	レコード情報へのポインタ
 *	@return	必ず0を返す
 */
static long del_record(dbCommon *pdbc)
{
	void *addr = pdbc->dpvt;

	pdbc->dpvt = NULL;
	free(addr);

	return 0L;
}

/**	initialize device processing
 *	@param[in]	after	before/after initializing record
 *						- @c 0 : before initializaing record
 *						- @c 1 : after initializing record
 *	@return	必ず0を返す
 */
static long init_intrx(int after)
{
	if (after == 0)
	{
		static struct dsxt dsxt = {add_record, del_record};

		devExtend(&dsxt);
	}
	return 0L;
}

/**	initialize specific record
 *	@param[in]	prec	レコード情報へのポインタ
 *	@return	必ず0を返す
 */
static long init_record(struct intrxRecord *prec)
{
	return 0L;
}

/**	get I/O interrupt info
 *	@param[in]		cmd		command@n
 *							- @c 0 add to scan list
 *							- @c 1 delete from scan list
 *	@param[in,out]	prec	pointer to intrxRecord structure
 *	@param[out]		ppvt	pointer to IOSCANPVT structure
 *	@retval	0				正常終了
 *	@retval	S_dev_NoInit	レコードの初期化に失敗している
 *
 *	- - -
 *	@see	drvRM5565_info, RM5565info_t
 */
static long get_ioint_info(int cmd, struct intrxRecord *prec, IOSCANPVT *ppvt)
{
	long status;

	if (prec->dpvt == NULL)
	{
		recGblRecordError(status = S_dev_NoInit, prec, __func__);
	}
	else
	{
		intrx_dpvt_t *dpvt = (intrx_dpvt_t *)(prec->dpvt);

		*ppvt = drvRM5565_info(dpvt->handle)->scan_nint[dpvt->intId];
		status = 0L;
	}
	return status;
}

/**	read record
 *	@param[in,out]	prec	pointer to intrxRecord structure
 *	@retval	0	正常終了
 *	@retval -1	ノードIDまたはNINTデータの取得エラー
 *
 *	- - -
 *	network interrupt dataが取得できなかった場合には、レコードをアラーム状態
 *	( _SOFT_ALARM_ / _INVALID_ALARM_ ) にする。@n
 *	node IDが取得できなかった場合には、レコードをアラーム状態
 *	( _SOFT_ALARM_ / _INVALID_ALARM_ ) にする。@n
 *	@see	drvRM5565_readIsd, drvRM5565_readSid
 */
static long read_record(struct intrxRecord *prec)
{
	intrx_dpvt_t *dpvt = (intrx_dpvt_t *)(prec->dpvt);
	epicsUInt32 data;
	epicsUInt8 nid;
	long status;

	assert(dpvt != NULL);

	status = 0L;

	/* get NINT data */
	if (drvRM5565_readISD(dpvt->handle, dpvt->intId, &data) != 0)
	{
		/* getting NINT data was failed */
		(void)errlogPrintf(__FILE__"(%d)[%s] ISD is empty\n", __LINE__, prec->name);
		recGblSetSevr(prec, SOFT_ALARM, INVALID_ALARM);
		status = -1L;
	}
	else
	{
		prec->val = data;
	}

	/* get node ID */
	if (drvRM5565_readSID(dpvt->handle, dpvt->intId, &nid) != 0)
	{
		/* getting node ID was failed */
		(void)errlogPrintf(__FILE__"(%d)[%s] SID is empty\n", __LINE__, prec->name);
		recGblSetSevr(prec, SOFT_ALARM, INVALID_ALARM);
		status = -1L;
	}
	else
	{
		prec->nid = nid;
	}
	return status;
}

/**	@cond		*/
static const struct {
	long number;
	DEVSUPFUN report;
	DEVSUPFUN init;
	DEVSUPFUN init_record;
	DEVSUPFUN get_ioint_info;
	DEVSUPFUN read_record;
} devIntrx5565 = {
	5,
	NULL,
	init_intrx,
	init_record,
	get_ioint_info,
	read_record
};
epicsExportAddress(dset, devIntrx5565);
/**	@endcond		*/

#if		0
#include <dbAccess.h>
#include <iocsh.h>
#include "drv5565priv.h"

/**	Simulate NINT receiving
 *	@param[in]	pvname	PV name
 *	@param[in]	node	source node ID
 *	@param[in]	data	NINT data
 *	@retval	0	OK
 *	@retval	-1	NG
 *
 *	- - -
 *	引数 __node__ と __data__ で指定されたパラメータで、
 *	引数 __pvname__ で示すレコードのネットワーク割込みをシミュレートする。@n
 *	引数 __pvname__ で示されるレコードは、DBlink でアクセス可能な _intrx_ レコードである必要がある。@n
 */
int intrxSimuNINT(const char *pvname, epicsUInt8 node, epicsUInt32 data)
{
	DBADDR addr;
	struct intrxRecord *prec;
	size_t intId;
	RM5565_t handle;
	RM5565priv_t *pRM;

	if ((pvname == NULL) || (*pvname == '\0'))
	{
		return -1;
	}
	if (dbNameToAddr(pvname, &addr) != 0)
	{
		(void)fprintf(stderr, "\"%s\" is external PV\n", pvname);
		return -1;
	}
	if (strcmp(addr.precord->rdes->name, "intrx") != 0)
	{
		(void)fprintf(stderr, "\"%s\" is not intrxRecord\n", pvname);
		return -1;
	}
	prec = (struct intrxRecord *)addr.precord;
	intId = (size_t)prec->inp.value.vmeio.signal - 1u;
	intId &= 3u;
	handle = drvRM5565_getHandle(prec->inp.value.vmeio.card);
	if (handle == NULL)
	{
		(void)fprintf(stderr, "\"%s\" is wrong setting\n", pvname);
		return -1;
	}
	pRM = (RM5565priv_t *)handle;
	epicsRingBytesPut(pRM->sid[intId], (char *)&node, sizeof(node));
	epicsRingBytesPut(pRM->isd[intId], (char *)&data, sizeof(data));
	if (dbProcess(addr.precord) != 0)
	{
		(void)fprintf(stderr, "dbProcess(\"%s\") failed\n", pvname);
		return -1;
	}
	return 0;
}
/**	@cond		*/
static const iocshArg intrxSimuNINT_arg0 = {"PV_name", iocshArgString};
static const iocshArg intrxSimuNINT_arg1 = {"source_node_id", iocshArgInt};
static const iocshArg intrxSimuNINT_arg2 = {"NINT_data", iocshArgInt};
static const iocshArg * const intrxSimuNINT_args[] = {
	&intrxSimuNINT_arg0,
	&intrxSimuNINT_arg1,
	&intrxSimuNINT_arg2
};
static const iocshFuncDef intrxSimuNINT_def = {"intrxSimuNINT", 3, intrxSimuNINT_args};
static void intrxSimuNINT_caller(const iocshArgBuf *args)
{
	if (intrxSimuNINT(args[0u].sval, args[1u].ival, args[2u].ival) != 0)
	{
		(void)fprintf(stderr, "intrxSimuNINT(\"%s\",%d,%d) failed\n",
		  args[0u].sval, args[1u].ival, args[2u].ival);
	}
}
static void intrxSimuNINT_registrar(void)
{
	iocshRegister(&intrxSimuNINT_def, intrxSimuNINT_caller);
}
epicsExportRegistrar(intrxSimuNINT_registrar);
/**	@endcond	*/

#endif

