/**
 *	@file	devInttx5565.c
 *	@brief	Device support for transmitting network interrupt of RM5565
 *	$Revision: 12 $
 *	$Date:: 2022-11-10 15:05:32 +0900#$
 *	$Author: iitsuka $
 *
 *	@page devsup_inttx Device Support for Transmitting Network-Interrupt
 *	Network-Interrupt 送信用デバイスサポート
 *
 *	このデバイスサポートは、@ref recsup_inttx "inttx"
 *	レコードタイプと組み合わせて使用する事で、
 *	Network-Interrupt の送信を行うためのものである。@n
 *	Device Type 名は @c "inttx5565"
 *
 *	@par Synopsis
 *	@code{.unparsed}
 *	record(inttx, "pv-name") {
 *		field(DTYP, "inttx5565")
 *		field(OUT,  "#C<card_number> S<int_type> @")
 *		field(NID,  "<target_node>")
 *		field(VAL,  "<data>")
 *	}
 *	@endcode
 *	@b card_number
 *	@li The card number designated in argument @e card in @ref RM5565Configure
 *
 *	@b int_type
 *	@li @c 1 ... Type 1 Network-Interrupt
 *	@li @c 2 ... Type 2 Network-Interrupt
 *	@li @c 3 ... Type 3 Network-Interrupt
 *	@li @c 4 ... Type 4 Network-Interrupt
 *
 *	@b target_node
 *	@li Network-Interrupt を送信する相手のノードID
 *
 *	@b data
 *	@li Network-Interrupt で送信するデータ
 *
 *	@par Usage
 *	@code{.unparsed}
 *	record(inttx, "RM5565:DEV0:INT1:TX") {
 *		field(DTYP, "inttx5565")
 *		field(OUT,  "#C0 S1 @")
 *			・
 *			・
 *			・
 *	}
 *	@endcode
 *	@c OUT フィールドでカード番号/Network-Interrupt Type を指定する。@n
 *	@c "C" に続く数字 (例では @c '0') でカード番号を、
 *	@c "S" に続く数字 (例では @c '1') で Network-Interrupt Type を指定しておく。@n
 *	@c NID フィールドに相手のノード ID を設定し、
 *	@code
 *	% caput RM5565:DEV0:INT1:TX.NID 1
 *	@endcode
 *	@c VAL フィールドに値を書き込むと
 *	@code
 *	% caput RM5565:DEV0:INT1:TX 10
 *	@endcode
 *	Network-Interrupt が送信される。@n
 *	なお、@c NID フィールドに 256 以上の値を設定すると、
 *	接続されている全ノードに送信する事が出来る。
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <alarm.h>
#include <dbScan.h>
#include <recSup.h>
#include <devSup.h>
#include <recGbl.h>
#include <epicsExport.h>
#include <cantProceed.h>
#include <errlog.h>

#include "inttxRecord.h"
#include "drv5565.h"

/**
 *	@brief	inttx レコード内部情報構造体
 */
typedef	struct
{
	RM5565_t handle;	/**< RM5565のハンドル									*/
	epicsUInt8 intId;	/**< レコードが対応するnetwork interruptのID (0 ~ 3)	*/
} inttx_dpvt_t;

/**	initialize record
 *	@param[in,out]	prec	pointer to inttxRecord structure
 *	@retval	0	OK
 *	@retval	-1	NG
 *
 *	- - -
 *	_OUT_ フィールドで指定されたカード番号/Network Interrupt Type に対応する __inttx__ レコードの初期化を行う。@n
 *	@code
 *	record(inttx, "pv-name") {
 *	        ・
 *	        ・
 *	    field(OUT, "#C0 S1 @")
 *	        ・
 *	        ・
 *	}
 *	@endcode
 *	@c "C" に続く数字 (ここでは @c '0') でカード番号が、
 *	@c "S" に続く数字 (ここでは @c '1') で Network Interrupt Type が指定される。
 *
 *	カード番号に対応するリフレクティブメモリのハンドル (@ref RM5565_t) を取得し、
 *	Network Interrupt Type とともに、@ref inttx_dpvt_t 構造体のそれぞれのメンバに格納する。@n
 *	値の格納された @ref inttx_dpvt_t は、_DPVT_ フィールドにセットされる。@n
 *	@see	drvRM5565_getHandle
 */
static long init_record(struct inttxRecord *prec)
{
	long status;

	if (prec->out.type != VME_IO)
	{
		(void)errlogPrintf(__FILE__"(%d) Bad OUT-type\n", __LINE__);
		status = S_dev_badOutType;
	}
	else
	{
		RM5565_t handle;

		handle = drvRM5565_getHandle(prec->out.value.vmeio.card);
		if (handle == NULL)
		{
			recGblRecordError(status = S_dev_badCard, prec, __func__);
		}
		else if ((1 > prec->out.value.vmeio.signal) || (prec->out.value.vmeio.signal > 4))
		{
			recGblRecordError(status = S_dev_badSignal, prec, __func__);
		}
		else
		{
			inttx_dpvt_t *dpvt = (inttx_dpvt_t *)mallocMustSucceed(sizeof(inttx_dpvt_t), __func__);
			dpvt->handle = handle;
			dpvt->intId = (epicsUInt8)prec->out.value.vmeio.signal - 1u;
			prec->dpvt = dpvt;
			status = 0L;
		}
	}
	return status;
}

/**	write record
 *	@param[in,out]	prec	pointer to inttxRecord structure
 *	@retval	0	OK
 *
 *	- - -
 *	_NID_ フィールドで指定されたノード ID を持つリフレクティブメモリに、
 *	network interruptを送信する。@n
 *	_NID_ フィールドが 256 以上だった場合には、全ノードに対して送信される。
 *	@see	drvRM5565_txInt
 */
static long write_record(struct inttxRecord *prec)
{
	long status;

	if (prec->dpvt == NULL)
	{
		status = -1L;
	}
	else
	{
		inttx_dpvt_t *dpvt = (inttx_dpvt_t *)(prec->dpvt);

		status = drvRM5565_txInt(dpvt->handle, prec->nid, dpvt->intId, prec->val);
	}
	return status;
}

/**	@cond		*/
struct {
	long number;
	DEVSUPFUN report;
	DEVSUPFUN init;
	DEVSUPFUN init_record;
	DEVSUPFUN get_ioint_info;
	DEVSUPFUN read_record;
} devInttx5565 = {
	5,
	NULL,
	NULL,
	init_record,
	NULL,
	write_record
};
epicsExportAddress(dset, devInttx5565);
/**	@endcond		*/

