/**
 *	@file	intrxRecord.c
 *	@brief	record process for receiving network interrupt of RM5565
 *	$Revision: 18 $
 *	$Date:: 2022-11-14 10:30:17 +0900#$
 *	$Author: iitsuka $
 *
 *	@page recsup_intrx Record Support for Receiving Network-Interrupt
 *	Network-Interrupt 受信用レコードサポート
 *
 *	このレコードタイプは、@ref devsup_intrx "intrx5565"
 *	デバイスサポートと組み合わせて使用する事で、
 *	Network-Interrupt の受信を行うためのものである。@n
 *
 *	@attention
 *	intrxRecord は、Network Interrupt Type(1～4)毎にインスタンスを 1 つずつしか作成できない。@n
 *	2 つ以上作成しようとすると意図しない動作となる。
 *
 *	@par Include
 *	@code
 *	#include <intrxRecord.h>
 *	@endcode
 *
 *	@par Field Summary
 *	Field   |Type       |DCT |Initial |Access |Modify |Rec Proc Monitor |PP
 *	:-----: |:--------- |:-- |:------ |:----- |:----- |:--------------- |:--
 *	@c INP  |INLINK     |Yes |0       |No     |No     |N/A              |No
 *	@c VAL  |LONG       |No  |0       |Yes    |No     |Yes              |No
 *	@c NID  |SHORT      |No  |0       |Yes    |No     |No               |No
 *	@c SNAM |STRING[41] |Yes |null    |Yes    |No     |No               |No
 *	@c ADEL |LONG       |Yes |0       |Yes    |Yes    |No               |No
 *	@c MDEL |LONG       |Yes |0       |Yes    |Yes    |No               |No
 *
 *	@par Field Description
 *	Name    |Summary          |Description
 *	:-----: |:--------------- |:-------------------------------
 *	@c INP  |Input Link       |"#C<card> S<int_type> @"@n card:カード番号@n int_type:Network Interrupt Type(1~4)
 *	@c VAL  |Value Field      |Network Interruptで通知されたデータがセットされる
 *	@c NID  |Node ID          |Network Interruptを送信したノードIDがセットされる
 *	@c SNAM |Subroutine Name  |必要であれば、Network Interrupt受信時に呼び出す関数名をセットする
 *	@c ADEL |Archive Deadband |same as other general records
 *	@c MDEL |Monitor Deadband |same as other general records
 *	SNAM フィールドには、レコードがプロセスされた時に実行する関数名を指定できる。指定しない事も可能。 @n
 *	関数の記述形式については下記の通り。@n
 *	@code{.c}
 *	void intFunc(struct intrxRecord *prec)
 *	{
 *					・
 *					・
 *					・
 *	}
 *	epicsRegisterFunction(intFunc);
 *	@endcode
 *	dbdファイルでの function 定義も併せて行う必要がある。
 *	@code
 *	function(intFunc)
 *	@endcode
 *
 *	@par Structure Definition
 *	@dontinclude intrxRecord.h
 *	@skip struct intrxRecord
 *	@until {
 *	@b 共通フィールド
 *	@until flnk
 *	@b 固有フィールド
 *	@until }
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <alarm.h>
#include <dbAccess.h>
#include <dbEvent.h>
#include <dbFldTypes.h>
#include <dbScan.h>
#include <devSup.h>
#include <recSup.h>
#include <recGbl.h>
#include <registryFunction.h>
#include <cantProceed.h>

/**	@cond		*/
#define	GEN_SIZE_OFFSET
/**	@endcond	*/
#include "intrxRecord.h"
#undef	GEN_SIZE_OFFSET

/**	structure of device support entry table.
 */
struct intrxdset
{
	long number;				/**< number of entries				*/
	DEVSUPFUN report;			/**< report routine					*/
	DEVSUPFUN init;				/**< initialize routine				*/
	DEVSUPFUN init_record;		/**< initialize record routine		*/
	DEVSUPFUN get_ioint_info;	/**< get I/O interrupt info routine	*/
	DEVSUPFUN read_record;		/**< read record routine			*/
};

typedef	void (*FUNC_t)(struct intrxRecord *);	/**< sub routine	*/

/**	check alarms
 *	@param[in,out]	prec	pointer to intrxRecord structure
 *
 *	- - -
 *	_UDF_ フィールドが __FALSE__ でない場合、
 *	レコードをアラーム状態 ( _UDF_ALARM_ / _INVALID_ALARM_ ) にする。
 */
static void checkAlarms(struct intrxRecord *prec)
{
	if (prec->udf != FALSE)
	{
		recGblSetSevr(prec, UDF_ALARM, INVALID_ALARM);
	}
}

/**	monitor process
 *	@param[in,out]	prec	pointer to intrxRecord structure
 */
static void monitor(struct intrxRecord *prec)
{
	epicsUInt16 monitor_mask = recGblResetAlarms(prec);
	epicsInt32 val, delta;

	val = prec->val;

	delta = (val > prec->mlst) ? (val - prec->mlst) : (prec->mlst - val);
	if (delta > prec->mdel)
	{
		monitor_mask |= DBE_VALUE;
		prec->mlst = val;
	}

	delta = (val > prec->alst) ? (val - prec->alst) : (prec->alst - val);
	if (delta > prec->adel)
	{
		monitor_mask |= DBE_LOG;
		prec->alst = val;
	}

	if (monitor_mask != 0u)
	{
		db_post_events(prec, &prec->val, monitor_mask);
	}
}

/**	read value
 *	@param[in,out]	prec	pointer to intrxRecord structure
 */
static long readValue(struct intrxRecord *prec)
{
	struct intrxdset *pdset = (struct intrxdset *)prec->dset;
	long status;

	if (prec->pact != FALSE)
	{
		status = (*pdset->read_record)(prec);
	}
	else
	{
		status = dbGetLink(&prec->siml, DBR_USHORT, &prec->simm, NULL, NULL);
		if (RTN_SUCCESS(status))
		{
			switch (prec->simm)
			{
			case menuYesNoNO:
				status = (*pdset->read_record)(prec);
				break;

			case menuYesNoYES:
				status = dbGetLink(&prec->siol, DBR_LONG, &prec->sval, NULL, NULL);
				if (RTN_SUCCESS(status))
				{
					prec->val = prec->sval;
					prec->udf = FALSE;
				}
				recGblSetSevr(prec, SIMM_ALARM, prec->sims);
				break;

			default:
				status = -1L;
				recGblSetSevr(prec, SOFT_ALARM, INVALID_ALARM);
				break;
			}
		}
	}
	return status;
}

/**	initialize record
 *	@param[in,out]	prec	pointer to intrxRecord structure
 *	@param[in]		pass
 *						- @c 0 :1st time call
 *						- @c 1 :2nd time call
 *	@retval	0		OK
 *	@retval	other	NG
 *
 *	- - -
 */
static long init_record(struct intrxRecord *prec, int pass)
{
	struct intrxdset *pdset;
	long status;
	FUNC_t pfunc;

	if (pass == 0)
	{
		return 0L;
	}

	if (prec->siml.type == CONSTANT)
	{
		recGblInitConstantLink(&prec->siml, DBF_USHORT, &prec->simm);
	}
	if (prec->siol.type == CONSTANT)
	{
		recGblInitConstantLink(&prec->siol, DBF_LONG, &prec->sval);
	}

	pfunc = (FUNC_t)registryFunctionFind(prec->snam);
	prec->pfnc = pfunc;

	pdset = (struct intrxdset *)prec->dset;
	if (pdset == NULL)
	{
		recGblRecordError(S_dev_noDSET, prec, "intrx: init_record");
		return S_dev_noDSET;
	}
	else if ((pdset->number < 5) || (pdset->read_record == NULL))
	{
		recGblRecordError(S_dev_missingSup, prec, "intrx: init_record");
		return S_dev_missingSup;
	}
	if (pdset->init_record != NULL)
	{
		status = (*pdset->init_record)(prec);
		if (RTN_SUCCESS(status) == FALSE)
		{
			return status;
		}
	}
	prec->alst = prec->mlst = prec->val;
	return 0L;
}

/**	process record
 *	@param[in,out]	prec	pointer to intrxRecord structure
 *	@retval	0		OK
 *	@retval	other	NG
 *
 *	- - -
 *	@see	readValue, checkAlarms, monitor
 */
static long process(struct intrxRecord *prec)
{
	struct intrxdset *pdset = (struct intrxdset *)prec->dset;
	long status;
	epicsUInt8 pact = prec->pact;

	if (pdset == NULL)
	{
		prec->pact = TRUE;
		recGblRecordError(S_dev_noDSET, prec, "intrx: process");
		return S_dev_noDSET;
	}
	else if ((pdset->number < 5) || (pdset->read_record == NULL))
	{
		prec->pact = TRUE;
		recGblRecordError(S_dev_missingSup, prec, "intrx: process");
		return S_dev_missingSup;
	}

	status = readValue(prec);
	if ((pact == FALSE) && (prec->pact != FALSE))
	{
		return 0L;
	}
	if (prec->pfnc != NULL)
	{
		(*((FUNC_t)prec->pfnc))(prec);
	}
	prec->pact = TRUE;

	recGblGetTimeStamp(prec);
	if (RTN_SUCCESS(status))
	{
		prec->udf = FALSE;
	}

	checkAlarms(prec);
	monitor(prec);
	recGblFwdLink(prec);

	prec->pact = FALSE;

	return status;
}

/**	@cond		*/
rset intrxRSET =
{
	RSETNUMBER,
	NULL,				/* report				*/
	NULL,				/* initialize			*/
	init_record,
	process,
	NULL,				/* special				*/
	NULL,				/* get_value			*/
	NULL,				/* cvt_dbaddr			*/
	NULL,				/* get_array_info		*/
	NULL,				/* put_array_info		*/
	NULL,				/* get_units			*/
	NULL,				/* get_precision		*/
	NULL,				/* get_enum_str			*/
	NULL,				/* get_enum_strs		*/
	NULL,				/* put_enum_str			*/
	NULL,				/* get_graphic_double	*/
	NULL,				/* get_control_double	*/
	NULL				/* get_alarm_double		*/
};
epicsExportAddress(rset, intrxRSET);
/**	@endcond	*/

