/**
 *	@file	inttxRecord.c
 *	@brief	record process for transmitting network interrupt of RM5565
 *	$Revision: 15 $
 *	$Date:: 2022-11-10 15:26:28 +0900#$
 *	$Author: iitsuka $
 *
 *	@page recsup_inttx Record Support for Transmitting Network-Interrupt
 *	Network-Interrupt 送信用レコードサポート
 *
 *	このレコードタイプは、@ref devsup_inttx "inttx5565"
 *	デバイスサポートと組み合わせて使用する事で、
 *	Network-Interrupt の送信を行うためのものである。@n
 *
 *	@par Include
 *	@code
 *	#include <inttxRecord.h>
 *	@endcode
 *
 *	@par Field Summary
 *	Field   |Type     |DCT |Initial |Access |Modify |Rec Proc Monitor |PP
 *	:-----: |:------- |:-- |:------ |:----- |:----- |:--------------- |:--
 *	@c OUT  |OUTLINK  |Yes |0       |No     |No     |N/A              |No
 *	@c VAL  |LONG     |Yes |0       |Yes    |Yes    |Yes              |Yes
 *	@c NID  |SHORT    |Yes |0       |Yes    |Yes    |No               |No
 *	@c DOL  |INLINK   |Yes |0       |No     |No     |N/A              |No
 *	@c OMSL |menuOmsl |Yes |0       |Yes    |Yes    |No               |No
 *	@c ADEL |LONG     |Yes |0       |Yes    |Yes    |No               |No
 *	@c MDEL |LONG     |Yes |0       |Yes    |Yes    |No               |No
 *
 *	@par Field Description
 *	Name    |Summary             |Description
 *	:-----: |:------------------ |:--------------------------------
 *	@c OUT  |Output Link         |"#C<card> S<int_type> @"@n card:カード番号@n int_type:Network Interrupt Type(1~4)
 *	@c VAL  |Value Field         |Network Interruptで通知するデータをセットする
 *	@c NID  |Node ID             |送信する相手のノードIDをセットする@n 256以上をセットすると全ノードに対して送信する
 *	@c DOL  |Desired Output Loc. |same as other general records
 *	@c OMSL |Output Mode Select  |same as other general records
 *	@c ADEL |Archive Deadband    |same as other general records
 *	@c MDEL |Monitor Deadband    |same as other general records
 *
 *	@par Structure Definition
 *	@dontinclude inttxRecord.h
 *	@skip struct inttxRecord
 *	@until {
 *	@b 共通フィールド
 *	@until flnk
 *	@b 固有フィールド
 *	@until }
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <alarm.h>
#include <dbAccess.h>
#include <dbEvent.h>
#include <dbScan.h>
#include <devSup.h>
#include <recSup.h>
#include <recGbl.h>
#include <cantProceed.h>

/**	@cond		*/
#define	GEN_SIZE_OFFSET
/**	@endcond	*/
#include "inttxRecord.h"
#undef	GEN_SIZE_OFFSET

/**	structure of device support entry table.
 */
struct inttxdset
{
	long number;				/**< number of entries				*/
	DEVSUPFUN report;			/**< report routine					*/
	DEVSUPFUN init;				/**< initialize routine				*/
	DEVSUPFUN init_record;		/**< initialize record routine		*/
	DEVSUPFUN get_ioint_info;	/**< get I/O interrupt info routine	*/
	DEVSUPFUN write_record;		/**< read record routine			*/
};

/**	check alarms
 *	@param[in,out]	prec	pointer to inttxRecord structure
 *
 *	- - -
 *	_UDF_ フィールドが __FALSE__ でない場合、
 *	レコードをアラーム状態 ( _UDF_ALARM_ / _INVALID_ALARM_ ) にする。
 */
static void checkAlarms(struct inttxRecord *prec)
{
	if (prec->udf != FALSE)
	{
		recGblSetSevr(prec, UDF_ALARM, INVALID_ALARM);
	}
}

/**	monitor process
 *	@param[in,out]	prec	pointer to inttxRecord structure
 */
static void monitor(struct inttxRecord *prec)
{
	epicsUInt16 monitor_mask = recGblResetAlarms(prec);
	epicsInt32 val, delta;

	val = prec->val;

	delta = (val > prec->mlst) ? (val - prec->mlst) : (prec->mlst - val);
	if (delta > prec->mdel)
	{
		monitor_mask |= DBE_VALUE;
		prec->mlst = val;
	}

	delta = (val > prec->alst) ? (val - prec->alst) : (prec->alst - val);
	if (delta > prec->adel)
	{
		monitor_mask |= DBE_LOG;
		prec->alst = val;
	}

	if (monitor_mask != 0u)
	{
		db_post_events(prec, &prec->val, monitor_mask);
	}
}

/**	write value
 *	@param[in,out]	prec	pointer to inttxRecord structure
 *	@retval	0		OK
 *	@retval	other	NG
 */
static long writeValue(struct inttxRecord *prec)
{
	struct inttxdset *pdset = (struct inttxdset *)(prec->dset);
	long status;

	if (prec->pact != FALSE)
	{
		status = (*pdset->write_record)(prec);
	}
	else
	{
		status = dbGetLink(&prec->siml, DBR_USHORT, &prec->simm, NULL, NULL);
		if (RTN_SUCCESS(status))
		{
			switch (prec->simm)
			{
			case menuYesNoNO:
				status = (*pdset->write_record)(prec);
				break;

			case menuYesNoYES:
				status = dbPutLink(&prec->siol, DBR_LONG, &prec->val, 1L);
				recGblSetSevr(prec, SIMM_ALARM, prec->sims);
				break;

			default:
				status = -1L;
				recGblSetSevr(prec, SOFT_ALARM, INVALID_ALARM);
				break;
			}
		}
	}
	return status;
}

/**	initialize record
 *	@param[in,out]	prec	pointer to inttxRecord structure
 *	@param[in]		pass	
 *							- @c 0 :1st time call
 *							- @c 1 :2nd time call
 *	@retval	0		OK
 *	@retval	other	NG
 */
static long init_record(struct inttxRecord *prec, int pass)
{
	struct inttxdset *pdset;
	long status;

	if (pass == 0)
	{
		return 0L;
	}

	if (prec->siml.type == CONSTANT)
	{
		recGblInitConstantLink(&prec->siml, DBF_USHORT, &prec->simm);
	}

	pdset = (struct inttxdset *)(prec->dset);
	if (pdset == NULL)
	{
		recGblRecordError(S_dev_noDSET, prec, "inttx: init_record");
		return S_dev_noDSET;
	}
	else if ((pdset->number < 5) || (pdset->write_record == NULL))
	{
		recGblRecordError(S_dev_missingSup, prec, "inttx: init_record");
		return S_dev_missingSup;
	}
	if (pdset->init_record != NULL)
	{
		status = (*pdset->init_record)(prec);
		if (RTN_SUCCESS(status) == FALSE)
		{
			return status;
		}
	}
	prec->alst = prec->mlst = prec->val;
	return 0L;
}

/**	process record
 *	@param[in,out]	prec	pointer to inttxRecord structure
 *	@retval	0		OK
 *	@retval	other	NG
 */
static long process(struct inttxRecord *prec)
{
	struct inttxdset *pdset = (struct inttxdset *)(prec->dset);
	long status;
	epicsInt32 val;
	epicsUInt8 pact = prec->pact;

	if (pdset == NULL)
	{
		prec->pact = TRUE;
		recGblRecordError(S_dev_noDSET, prec, "inttx: write_record");
		return S_dev_noDSET;
	}
	else if ((pdset->number < 5) || (pdset->write_record == NULL))
	{
		prec->pact = TRUE;
		recGblRecordError(S_dev_missingSup, prec, "inttx: write_record");
		return S_dev_missingSup;
	}
	if (prec->pact == FALSE)
	{
		if ((prec->dol.type != CONSTANT) && (prec->omsl == menuOmslclosed_loop))
		{
			if (RTN_SUCCESS(dbGetLink(&prec->dol, DBR_LONG, &val, NULL, NULL)))
			{
				prec->val = val;
				prec->udf = FALSE;
			}
		}
	}
	checkAlarms(prec);
	if (prec->nsev < INVALID_ALARM)
	{
		status = writeValue(prec);
	}
	else
	{
		status = 0L;
	}
	if ((pact == FALSE) && (prec->pact != FALSE))
	{
		return 0L;
	}
	prec->pact = TRUE;

	recGblGetTimeStamp(prec);
	monitor(prec);
	recGblFwdLink(prec);
	prec->pact = FALSE;

	return status;
}

/**	@cond		*/
rset inttxRSET =
{
	RSETNUMBER,
	NULL,				/* report				*/
	NULL,				/* initialize			*/
	init_record,
	process,
	NULL,				/* special				*/
	NULL,				/* get_value			*/
	NULL,				/* cvt_dbaddr			*/
	NULL,				/* get_array_info		*/
	NULL,				/* put_array_info		*/
	NULL,				/* get_units			*/
	NULL,				/* get_precision		*/
	NULL,				/* get_enum_str			*/
	NULL,				/* get_enum_strs		*/
	NULL,				/* put_enum_str			*/
	NULL,				/* get_graphic_double	*/
	NULL,				/* get_control_double	*/
	NULL				/* get_alarm_double		*/
};
epicsExportAddress(rset, inttxRSET);
/**	@endcond	*/

